@@ -1,12 +1,13 @@
 import json
 import pathlib
 import re
+import weakref
 from abc import ABC, abstractmethod
 from collections.abc import MutableMapping, Hashable
 from copy import deepcopy
 from enum import Enum
 from pathlib import Path, PureWindowsPath, PurePath, PurePosixPath
-from typing import Iterator, Optional, Any, List, Dict, Union
+from typing import Iterator, Optional, Any, List, Dict, Union, TYPE_CHECKING
 
 from paquo._base import QuPathBase
 from paquo._logging import redirect, get_logger
@@ -14,10 +15,34 @@ from paquo._utils import cached_property
 from paquo.hierarchy import QuPathPathObjectHierarchy
 from paquo.java import String, DefaultProjectImageEntry, ImageType, ImageData, IOException, URI, URISyntaxException, \
     PathIO, File, BufferedImage, FileNotFoundException
+if TYPE_CHECKING:
+    import paquo.projects
 
 _log = get_logger(__name__)
 
 
+# [URI:java-python]
+# NOTE: pathlib handles URIs a little different to QuPath's java URIs
+#   having looked into it a little bit it seems neither are entirely
+#   rfc3986 compliant because they both try to be permissive with
+#   broken URIs...
+#   For the sake of moving forward we go with the workarounds below.
+#   This should all be replaced with rfc3986 compliant URI handling.
+def _normalize_pathlib_uris(uri):
+    """this will correctly unescape and normalize uri's received from pathlib.Path.as_uri()"""
+    # https://docs.oracle.com/javase/7/docs/api/java/net/URI.html section Identities
+    u = URI(uri)
+    return URI(
+        u.getScheme(),
+        u.getUserInfo(),
+        u.getHost(),
+        u.getPort(),
+        u.getPath(),
+        u.getQuery(),
+        u.getFragment()
+    )
+
+
 class ImageProvider(ABC):
     """Maps image ids to paths and paths to image ids."""
 
@@ -45,19 +70,22 @@ class ImageProvider(ABC):
         """
         Parses an URI representing a file system path into a Path.
         """
+        # TODO: needs way more tests... See note [URI:java-python]
         try:
             java_uri = URI(uri)
         except URISyntaxException:
             raise ValueError(f"not a valid uri '{uri}'")
-
+        # test current scheme support
         if str(java_uri.getScheme()) != "file":
             raise NotImplementedError("paquo only supports file:/ URIs as of now")
-        path_str = str(java_uri.getPath())
-
+        else:
+            path_str = str(java_uri.getPath())
+        # fixme: this should be replaced with something more reliable...
         # check if we encode a windows path
-        # fixme: not sure if there's a better way to do this...
-        if re.match("/[A-Z]:/", path_str):
+        if re.match(r"/[A-Z]:/[^/]", path_str):
             return PureWindowsPath(path_str[1:])
+        elif re.match(r"//(?P<share>[^/]+)/(?P<directory>[^/]+)/", path_str):
+            return PureWindowsPath(path_str)
         else:
             return PurePosixPath(path_str)
 
@@ -66,20 +94,29 @@ class ImageProvider(ABC):
         """
         Convert a python path object to an URI
         """
+        # TODO: needs way more tests... See note [URI:java-python]
         if not path.is_absolute():
             raise ValueError("uri_from_path requires an absolute path")
-        return str(URI(path.as_uri()).toString())
+        java_uri = str(_normalize_pathlib_uris(path.as_uri()).toString())
+        # fixme: this should be replaced with a rfc3896 compliant solution...
+        if re.match("file://([^/]|$)", java_uri):
+            uri = f"file:////{java_uri[7:]}"  # network shares have redundant authority on the java side
+        # vvv this would only be required if we wouldn't normalize the uri like above
+        # elif re.match("file:///([^/]|$)", java_uri):
+        #     uri = f"file:/{java_uri[8:]}"  # the local windows absolute paths don't
+        else:
+            uri = java_uri
+        return uri
 
     @staticmethod
     def compare_uris(a: str, b: str) -> bool:
-        # ... comma encoding is problematic
-        # python url encodes commas, but java doesn't
-        # need to add more tests
-        uri_a = URI(a)
-        uri_b = URI(b)
-        if any(str(uri.getScheme()) != "file" for uri in [uri_a, uri_b]):
-            raise NotImplementedError("currently untested ...")
-        return bool(uri_a.getPath() == uri_b.getPath())
+        """
+        Test if two URIs point two the same resource
+        """
+        # TODO: needs way more tests... See note [URI:java-python]
+        uri_a = _normalize_pathlib_uris(a)
+        uri_b = _normalize_pathlib_uris(b)
+        return bool(uri_a.equals(uri_b))
 
     @classmethod
     def __subclasshook__(cls, C):
@@ -202,10 +239,14 @@ class _RecoveredReadOnlyImageServer:
 class _ProjectImageEntryMetadata(MutableMapping):
     """provides a python dict interface for image entry metadata"""
 
-    def __init__(self, entry: DefaultProjectImageEntry) -> None:
-        self._entry = entry
+    def __init__(self, image: 'QuPathProjectImageEntry') -> None:
+        self._image = image
+        self._entry = image.java_object
 
     def __setitem__(self, k: str, v: str) -> None:
+        # noinspection PyProtectedMember
+        if self._image._readonly:
+            raise AttributeError("project in readonly mode")
         if not isinstance(k, str):
             raise TypeError(f"key must be of type `str` got `{type(k)}`")
         if not isinstance(v, str):
@@ -213,6 +254,9 @@ class _ProjectImageEntryMetadata(MutableMapping):
         self._entry.putMetadataValue(String(str(k)), String(str(v)))
 
     def __delitem__(self, k: str) -> None:
+        # noinspection PyProtectedMember
+        if self._image._readonly:
+            raise AttributeError("project in readonly mode")
         if not isinstance(k, str):
             raise TypeError(f"key must be of type `str` got `{type(k)}`")
         self._entry.removeMetadataValue(String(str(k)))
@@ -235,6 +279,9 @@ class _ProjectImageEntryMetadata(MutableMapping):
         return bool(self._entry.containsMetadata(String(str(item))))
 
     def clear(self) -> None:
+        # noinspection PyProtectedMember
+        if self._image._readonly:
+            raise AttributeError("project in readonly mode")
         self._entry.clearMetadata()
 
     def __repr__(self):
@@ -244,15 +291,23 @@ class _ProjectImageEntryMetadata(MutableMapping):
 class _ImageDataProperties(MutableMapping):
     """provides a python dict interface for image data properties"""
 
-    def __init__(self, image_data: ImageData) -> None:
-        self._image_data = image_data
+    def __init__(self, image: 'QuPathProjectImageEntry') -> None:
+        self._image = image
+        # noinspection PyProtectedMember
+        self._image_data = image._image_data
 
     def __setitem__(self, k: str, v: Any) -> None:
+        # noinspection PyProtectedMember
+        if self._image._readonly:
+            raise AttributeError("project in readonly mode")
         if not isinstance(k, str):
             raise TypeError(f"key must be of type `str` got `{type(k)}`")
         self._image_data.setProperty(String(k), v)
 
     def __delitem__(self, k: str) -> None:
+        # noinspection PyProtectedMember
+        if self._image._readonly:
+            raise AttributeError("project in readonly mode")
         if not isinstance(k, str):
             raise TypeError(f"key must be of type `str` got `{type(k)}`")
         self._image_data.removeProperty(String(k))
@@ -319,7 +374,8 @@ class QuPathImageType(str, Enum):
 
 class QuPathProjectImageEntry(QuPathBase[DefaultProjectImageEntry]):
 
-    def __init__(self, entry: DefaultProjectImageEntry) -> None:
+    def __init__(self, entry: DefaultProjectImageEntry,
+                 *, _project_ref: Optional['paquo.projects.QuPathProject'] = None) -> None:
         """Wrapper for qupath image entries
 
         this is normally not instantiated by the user
@@ -327,7 +383,13 @@ class QuPathProjectImageEntry(QuPathBase[DefaultProjectImageEntry]):
         if not isinstance(entry, DefaultProjectImageEntry):
             raise ValueError("don't instantiate directly. use `QuPathProject.add_image`")
         super().__init__(entry)
-        self._metadata = _ProjectImageEntryMetadata(entry)
+        self._project_ref = weakref.ref(_project_ref) if _project_ref else lambda: None
+        self._metadata = _ProjectImageEntryMetadata(self)
+
+    @property
+    def _readonly(self):
+        p = self._project_ref()
+        return getattr(p, "_readonly", False) if p else True
 
     @cached_property
     def _image_data(self):
@@ -348,7 +410,7 @@ class QuPathProjectImageEntry(QuPathBase[DefaultProjectImageEntry]):
 
     @cached_property
     def _properties(self):
-        return _ImageDataProperties(self._image_data)
+        return _ImageDataProperties(self)
 
     @cached_property
     def _image_server(self):
@@ -375,6 +437,8 @@ class QuPathProjectImageEntry(QuPathBase[DefaultProjectImageEntry]):
 
     @image_name.setter
     def image_name(self, name: str) -> None:
+        if self._readonly:
+            raise AttributeError("project in readonly mode")
         self.java_object.setImageName(String(name))
 
     # remove until there's a good use case for this...
@@ -391,6 +455,8 @@ class QuPathProjectImageEntry(QuPathBase[DefaultProjectImageEntry]):
 
     @image_type.setter
     def image_type(self, value: QuPathImageType) -> None:
+        if self._readonly:
+            raise AttributeError("project in readonly mode")
         if not isinstance(value, QuPathImageType):
             raise TypeError("requires a QuPathImageType enum")
         self._image_data.setImageType(value.java_enum)
@@ -405,30 +471,44 @@ class QuPathProjectImageEntry(QuPathBase[DefaultProjectImageEntry]):
 
     @description.setter
     def description(self, text: str) -> None:
+        if self._readonly:
+            raise AttributeError("project in readonly mode")
         self.java_object.setDescription(text)
 
     @property
     def width(self):
+        """image width in pixels"""
         return int(self._image_server.getWidth())
 
     @property
     def height(self):
+        """image height in pixels"""
         return int(self._image_server.getHeight())
 
     @property
     def num_channels(self):
+        """number of channels in the image"""
         return int(self._image_server.nChannels())
 
     @property
     def num_z_slices(self):
+        """number of z_slices in the image"""
         return int(self._image_server.nZSlices())
 
     @property
     def num_timepoints(self):
+        """number of time points in the image"""
         return int(self._image_server.nTimepoints())
 
     @cached_property
     def downsample_levels(self) -> List[Dict[str, float]]:
+        """downsample levels provided by the image
+
+        Notes
+        -----
+        The available downsample levels can differ dependent
+        on which image backend is used by QuPath
+        """
         md = self._image_server.getMetadata()
         levels = []
         for level in range(int(md.nLevels())):
@@ -447,6 +527,8 @@ class QuPathProjectImageEntry(QuPathBase[DefaultProjectImageEntry]):
 
     @metadata.setter
     def metadata(self, value: dict) -> None:
+        if self._readonly:
+            raise AttributeError("project in readonly mode")
         self._metadata.clear()
         self._metadata.update(value)
 
@@ -457,6 +539,8 @@ class QuPathProjectImageEntry(QuPathBase[DefaultProjectImageEntry]):
 
     @properties.setter
     def properties(self, value):
+        if self._readonly:
+            raise AttributeError("project in readonly mode")
         self._properties.clear()
         self._properties.update(value)
 
@@ -464,13 +548,79 @@ class QuPathProjectImageEntry(QuPathBase[DefaultProjectImageEntry]):
     def hierarchy(self) -> QuPathPathObjectHierarchy:
         """the image entry hierarchy. it contains all annotations"""
         try:
-            return QuPathPathObjectHierarchy(self._image_data.getHierarchy())
+            return QuPathPathObjectHierarchy(self._image_data.getHierarchy(), _image_ref=self)
         except OSError:
             _log.warning("could not open image data. loading annotation hierarchy from project.")
-            return QuPathPathObjectHierarchy(self.java_object.readHierarchy())
+            return QuPathPathObjectHierarchy(self.java_object.readHierarchy(), _image_ref=self)
 
     def __repr__(self):
-        return f"<ImageEntry('{self.image_name}')>"
+        return f"<ImageEntry image_name='{self.image_name}'>"
+
+    def _repr_html_(self, compact=False, index=0):
+        from base64 import b64encode
+        from paquo._repr import br, h4, img, div, p, span
+
+        img_css = {
+            "max-width": "100px",
+            "max-height": "100px",
+            "border": "1px solid",
+            "margin": "auto",
+        }
+        header_css = {
+            "position": "absolute",
+            "top": "-1.6em",
+            "width": "100px",
+            "overflow": "hidden",
+            "text-overflow": "ellipsis",
+            "font-size": "0.75em",
+        }
+        container_css = {
+            "display": "flex",
+            "align-items": "center",
+            "justify-content": "center",
+            "position": "relative",
+            "width": "100px",
+            "height": "100px",
+            "background": "#ddd",
+            "margin": "2px",
+        }
+
+        try:
+            with (self.entry_path / "thumbnail.jpg").open(mode="rb") as f:
+                data = b64encode(f.read()).decode('utf-8')
+        except FileNotFoundError:  # pragma: no cover
+            image = span(style={"font-size": "3em"}, text="?")
+        else:
+            image = img(title=self.image_name,
+                        src=f"data:image/jpeg;base64,{data}",
+                        style=img_css)
+        if compact:
+            container_css["margin-top"] = "1em"
+            return div(
+                span(text=f"[{index}]\xa0{self.image_name}", style=header_css),
+                image,
+                style=container_css
+            )
+
+        try:
+            uri = self.uri[5:]
+        except RuntimeError as err:  # pragma: no cover
+            uri = f"N/A ({err})"
+        return div(
+            h4(text=f"Image: {self.image_name}", style={"margin-top": "0"}),
+            p(
+                span(text="path: ", style={"font-weight": "bold"}),
+                span(text=uri),
+                br(),
+                span(text="type: ", style={"font-weight": "bold"}),
+                span(text=str(self.image_type.value)),
+                style={"margin": "0.5em"},
+            ),
+            div(
+                image,
+                style=container_css,
+            )
+        )
 
     @property
     def uri(self):
@@ -501,5 +651,13 @@ class QuPathProjectImageEntry(QuPathBase[DefaultProjectImageEntry]):
     def save(self):
         """save image entry"""
         with redirect(stdout=True, stderr=True):
-            if self.is_readable() and self.is_changed():
-                self.java_object.saveImageData(self._image_data)
+            if self._readonly:
+                raise IOError("project in readonly mode")
+            if self.is_readable():
+                if self.is_changed():
+                    self.java_object.saveImageData(self._image_data)
+                    _log.info(f"Changes saved for '{self.image_name}'")
+                else:
+                    _log.info(f"Saving skipped for '{self.image_name}': no changes")
+            else:
+                _log.warning(f"Saving skipped for '{self.image_name}': uri '{self.uri}' not reachable")
