@@ -1,7 +1,8 @@
 import platform
 import shutil
 import tempfile
-from pathlib import Path
+from operator import itemgetter
+from pathlib import Path, PureWindowsPath, PurePosixPath
 
 import pytest
 
@@ -14,7 +15,7 @@ from paquo.projects import QuPathProject
 @pytest.fixture(scope='module')
 def image_entry(svs_small):
     with tempfile.TemporaryDirectory(prefix='paquo-') as tmpdir:
-        qp = QuPathProject(tmpdir)
+        qp = QuPathProject(tmpdir, mode='x')
         entry = qp.add_image(svs_small)
         yield entry
 
@@ -30,7 +31,7 @@ def removable_svs_small(svs_small):
 @pytest.fixture(scope='function')
 def project_with_removed_image(removable_svs_small):
     with tempfile.TemporaryDirectory(prefix='paquo-') as tmpdir:
-        qp = QuPathProject(tmpdir)
+        qp = QuPathProject(tmpdir, mode='x')
         _ = qp.add_image(removable_svs_small, image_type=QuPathImageType.BRIGHTFIELD_H_E)
         qp.save()
         removable_svs_small.unlink()
@@ -40,7 +41,7 @@ def project_with_removed_image(removable_svs_small):
 @pytest.fixture(scope='function')
 def project_with_removed_image_without_image_data(removable_svs_small):
     with tempfile.TemporaryDirectory(prefix='paquo-') as tmpdir:
-        qp = QuPathProject(tmpdir)
+        qp = QuPathProject(tmpdir, mode='x')
         _ = qp.add_image(removable_svs_small)
         qp.save()
         removable_svs_small.unlink()
@@ -150,7 +151,7 @@ def test_readonly_recovery_hierarchy(project_with_removed_image_without_image_da
 
 
 def test_readonly_recovery_image_server(project_with_removed_image):
-    with QuPathProject(project_with_removed_image, mode='r+') as qp:
+    with QuPathProject(project_with_removed_image, mode='r') as qp:
         image_entry = qp.images[0]
 
         assert image_entry.height
@@ -209,39 +210,101 @@ def test_image_type(image_entry):
     assert image_entry.image_type == QuPathImageType.BRIGHTFIELD_H_E
 
 
+TEST_URIS = {
+    "no-uri": {  # weird non-uri found in one project
+        "uri": "\\\\SHARE\\site\\2020-01-01\\image 123-X,X.svs",
+        "parts": None,
+        "path": None,
+        "exception": ValueError,
+    },
+    "network-share": {  # share on win
+        "uri": "file:////SHARE/site/ABC/image.svs",
+        "parts": ('ABC', 'image.svs'),
+        "path": PureWindowsPath("//SHARE/site/", 'ABC', 'image.svs'),
+        "exception": None,
+    },
+    "win-simple": {  # win-style uris
+        "uri": "file:/C:/ABC/image.svs",
+        "parts": ('ABC', 'image.svs'),
+        "path": PureWindowsPath("C:/", 'ABC', 'image.svs'),
+        "exception": None,
+    },
+    "win-spaces": {
+        "uri": "file:/C:/ABC%20-%20ABC/image.svs",
+        "parts": ('ABC - ABC', 'image.svs'),
+        "path": PureWindowsPath("C:/", 'ABC - ABC', 'image.svs'),
+        "exception": None,
+    },
+    "win-commas": {
+        "uri": "file:/C:/ABC,ABC/image.svs",
+        "parts": ('ABC,ABC', 'image.svs'),
+        "path": PureWindowsPath("C:/", 'ABC,ABC', 'image.svs'),
+        "exception": None,
+    },
+    "win-space-comma": {
+        "uri": "file:/C:/ABC%20,%20ABC/image.svs",
+        "parts": ('ABC , ABC', 'image.svs'),
+        "path": PureWindowsPath("C:/", 'ABC , ABC', 'image.svs'),
+        "exception": None,
+    },
+    "win-comb-01": {
+        "uri": "file:/D:/2020-01-01/image.svs",
+        "parts": ('2020-01-01', 'image.svs'),
+        "path": PureWindowsPath("D:/", '2020-01-01', 'image.svs'),
+        "exception": None,
+    },
+    "win-comb-02": {
+        "uri": "file:/F:/2020-01-01/image-123-abc.svs",
+        "parts": ('2020-01-01', 'image-123-abc.svs'),
+        "path": PureWindowsPath("F:/", '2020-01-01', 'image-123-abc.svs'),
+        "exception": None,
+    },
+    "win-comb-03": {
+        "uri": "file:/U:/site/ABC%20ABC/image.svs",
+        "parts": ('site', 'ABC ABC', 'image.svs'),
+        "path": PureWindowsPath("U:/", 'site', 'ABC ABC', 'image.svs'),
+        "exception": None,
+    },
+    "posix": {  # posix-style uri
+        "uri": "file:/site/image.svs",
+        "parts": ('site', 'image.svs'),
+        "path": PurePosixPath("/", 'site', 'image.svs'),
+        "exception": None,
+    },
+}
+
+
 @pytest.mark.parametrize(
-    "uri,parts1n,exc", [
-        ("\\\\SHARE\\site\\2020-01-01\\image 123-X,X.svs", None, ValueError),  # weird non-uri found in one project
-        ("file:////SHARE/site/image.svs", ('SHARE', 'site', 'image.svs'), None),  # share on win
-        ("file:/C:/ABC/image.svs", ('ABC', 'image.svs'), None),  # win-style uris
-        ("file:/C:/ABC%20-%20ABC/image.svs", ('ABC - ABC', 'image.svs'), None),
-        ("file:/C:/ABC,ABC/image.svs", ('ABC,ABC', 'image.svs'), None),
-        ("file:/C:/ABC%20,%20ABC/image.svs", ('ABC , ABC', 'image.svs'), None),
-        ("file:/D:/2020-01-01/image.svs", ('2020-01-01', 'image.svs'), None),
-        ("file:/F:/2020-01-01/image-123-abc.svs", ('2020-01-01', 'image-123-abc.svs'), None),
-        ("file:/U:/site/ABC%20ABC/image.svs", ('site', 'ABC ABC', 'image.svs'), None),
-        ("file:/site/image.svs", ('site', 'image.svs'), None),  # posix-style uri
-    ],
-    ids=[
-        "no-uri",
-        "network-share",
-        "win-simple",
-        "win-spaces",
-        "win-commas",
-        "win-space-comma",
-        "win-comb-01",
-        "win-comb-02",
-        "win-comb-03",
-        "posix",
-    ]
+    "uri,exc", list(map(itemgetter("uri", "exception"), TEST_URIS.values())),
+    ids=list(TEST_URIS.keys())
 )
-def test_image_provider_path_from_uri(uri, parts1n, exc):
+def test_image_provider_raise_if_invalid_uri(uri, exc):
     cm = pytest.raises(exc) if exc else nullcontext()
     with cm:
-        path = ImageProvider.path_from_uri(uri)
-        assert path.parts[1:] == parts1n
-        new_uri = ImageProvider.uri_from_path(path)
-        assert ImageProvider.compare_uris(uri, new_uri)
+        ImageProvider.path_from_uri(uri)
+
+
+@pytest.mark.parametrize(
+    "uri,path", [(x["uri"], x["path"]) for x in TEST_URIS.values() if x["exception"] is None],
+    ids=[k for k, x in TEST_URIS.items() if x["exception"] is None]
+)
+def test_image_provider_uri_from_path(uri, path):
+    # noinspection PyTypeChecker
+    new_uri = ImageProvider.uri_from_path(path)
+    assert uri == new_uri
+    # NOTE: the following test is of course weaker
+    assert ImageProvider.compare_uris(uri, new_uri)
+
+
+@pytest.mark.parametrize(
+    "uri,path", [(x["uri"], x["path"]) for x in TEST_URIS.values() if x["exception"] is None],
+    ids=[k for k, x in TEST_URIS.items() if x["exception"] is None]
+)
+def test_image_provider_path_from_uri(uri: str, path: Path):
+    c_path = ImageProvider.path_from_uri(uri)
+    assert c_path.is_absolute()
+    assert type(c_path) is type(path)
+    assert c_path.parts == path.parts
 
 
 def test_image_provider_ducktyping():
@@ -275,7 +338,7 @@ def test_image_provider_default_implementation():
     assert set(ip.rebase('file:/abc.svs', 'file:/efg.svs')) == {None}
 
 
-def test_image_provider_uri_from_path():
+def test_image_provider_uri_from_relpath_and_abspath():
     with pytest.raises(ValueError):
         ImageProvider.uri_from_path(Path('./abc.svs'))
 
